from pandas import Series, Timedelta, TimedeltaIndex, to_datetime, to_timedelta
import numpy as np

origin = to_datetime("2000-1-1")


def check_binop_timezones(first, second):
    assert first.tz == second.tz, "operands must have the same timezone, or none at all"


def _convert_date_to_float(val, tz=None):
    if val is None:
        return None
    if hasattr(val, "__iter__"):
        if not isinstance(val, Series):
            val = Series(val)
        deltas = TimedeltaIndex(val - origin.tz_localize(tz))
        return list(deltas / Timedelta(1, "h"))
    return (val - origin.tz_localize(tz)) / Timedelta(1, "h")


def _convert_float_to_date(val, tz=None):
    if hasattr(val, "__iter__"):
        if not isinstance(val, np.ndarray):
            val = np.array(val)
        return list(to_timedelta(val * 3600, unit="s") + origin.tz_localize(tz))
    return to_timedelta(val * 3600, unit="s") + origin.tz_localize(tz)



from staircase.core.stats.statistic import (
    _get_integral_and_mean,
    get_integral_and_mean,
    integrate,
    mean,
    median,
    mode,
    percentile,
    std,
    var,
)

from staircase.core.stats.distribution import (
    ecdf_stairs,
    hist,
    hist_from_ecdf,
    percentile_stairs,
    percentile_Stairs,
)


def add_operations(cls):

    cls._get_integral_and_mean = _get_integral_and_mean
    cls.get_integral_and_mean = get_integral_and_mean
    cls.integrate = integrate
    cls.mean = mean
    cls.median = median
    cls.mode = mode
    cls.percentile = percentile
    cls.std = std
    cls.var = var

    cls.ecdf_stairs = ecdf_stairs
    cls.hist = hist
    cls.hist_from_ecdf = hist_from_ecdf
    cls.percentile_stairs = percentile_stairs
    cls.percentile_Stairs = percentile_Stairs



var_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()

.. plot::
    :context: close-figs

    >>> s1.var()
    0.6875

    >>> s1.var(lower=0, upper=6)
    0.4722222222222224
"""

std_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()

.. plot::
    :context: close-figs

    >>> s1.std()
    0.82915619758885

    >>> s1.std(lower=0, upper=6)
    0.6871842709362769
"""

integral_and_mean_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.get_integral_and_mean(3, 4.5)
    (0.5, 0.3333333333333333)
"""

integrate_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.integrate(3, 4.5)
    0.5
"""

mean_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.mean(3, 4.5)
    0.3333333333333333
"""

mode_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s2.plot()
    >>> s2.mode()
    -1.0
    >>> s2.mode(0,3)
    0.5
"""


median_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.median()
    0.5
    >>> s1.median(2,5)
    0.0
"""

percentile_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.percentile(70)
    1.0
    >>> s1.percentile(25)
    -0.5
    >>> s1.percentile(40, lower=1.5, upper=3.5)
    0.0
"""

percentile_stairs_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12,5))
    >>> s2.plot(axes[0])
    >>> axes[0].set_title("s2")
    >>> s2_percentiles = s2.percentile_stairs()
    >>> s2_percentiles.plot(axes[1])
    >>> axes[0].set_title("s2 percentiles")
    >>> s2_percentiles(55)
    0.0
    >>> s2_percentiles(75)
    0.5
"""

hist_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()

.. plot::
    :context: close-figs

    >>> s1.hist()
    [-1, 0)    0.25
    [0, 1)     0.25
    [1, 2)     0.50
    dtype: float64

    >>> s1.hist(closed='right')
    (-2, -1]    0.25
    (-1, 0]     0.25
    (0, 1]      0.50
    dtype: float64

    >>> s1.hist(2, 4.5)
    [-1, 0)    0.2
    [0, 1)     0.4
    [1, 2)     0.4
    dtype: float64

    >>> s1.hist(bin_edges=(-1,1,3))
    [-1, 1)    0.5
    [1, 3)     0.5
    dtype: float64

    >>> s1.hist(bin_edges=(-1, 1))
    [-1, 1)    0.5
    dtype: float64
"""

ecdf_stairs_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s2.plot()

.. plot::
    :context: close-figs

    >>> ecdf = s2.ecdf_stairs(1,5)
    >>> ecdf.plot()
    >>> plt.show()


    >>> print(f'{ecdf(0):.2%} of values, for s2 between 1 and 5, are less than or equal to 0')
    75.00% of values, for s2 between 1 and 5, are less than or equal to 0

    >>> print(f'{ecdf(0, how="left"):.2%} of values, for s2 between 1 and 5, are strictly less than 0')
    50.00% of values, for s2 between 1 and 5, are strictly less than 0

    >>> print(f'{ecdf(0.2) - ecdf(-1):.2%} of values, for s2 between 1 and 5, are in (-1, 0.2]')
    25.00% of values, for s2 between 1 and 5, are in (-1, 0.2]

    >>> print(f'{ecdf(0.2, how="left") - ecdf(-1, how="left"):.2%} of values, for s2 between 1 and 5, are in [-1, 0.2)')
    75.00% of values, for s2 between 1 and 5, are in [-1, 0.2)
"""

hist_from_ecdf_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> import staircase as sc
    >>> s1_ecdf_stairs = s1.ecdf_stairs()
    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(8,5), sharey=False, sharex=False)
    >>> for ax, title, stair_instance in zip(axes, ("s1", "s1 ecdf"), (s1, s1_ecdf_stairs)):
    ...     stair_instance.plot(ax)
    ...     ax.set_title(title)
    >>> plt.show()

    >>> sc.hist_from_ecdf(s1_ecdf_stairs)
    [-1, 0)    0.25
    [0, 1)     0.25
    [1, 2)     0.50
    dtype: float64

    >>> sc.hist_from_ecdf(s1_ecdf_stairs, bin_edges=(-1,1,3))
    [-1, 1)    0.5
    [1, 3)     0.5
    dtype: float64
"""



_symbol_map = {
    "addition": "+",
    "subtraction": "-",
    "multiplication": r"*",
    "division": "/",
    "lt": "<",
    "le": "<=",
    "gt": ">",
    "ge": ">=",
    "eq": "==",
    "ne": "!=",
    "and": "&",
    "or": "|",
    "negate": "-",
    "invert": "~",
}

_see_also_map = {
    "addition": "Stairs.subtract",
    "subtraction": "Stairs.add",
    "multiplication": "Stairs.divide",
    "division": "Stairs.multiply",
    "lt": "Stairs.gt, Stairs.le, Stairs.ge",
    "le": "Stairs.lt, Stairs.gt, Stairs.ge",
    "gt": "Stairs.lt, Stairs.le, Stairs.ge",
    "ge": "Stairs.lt, Stairs.gt, Stairs.le",
    "eq": "Stairs.ne, Stairs.identical",
    "ne": "Stairs.eq, Stairs.identical",
    "and": "Stairs.logical_or",
    "or": "Stairs.logical_and",
    "negate": "Stairs.subtract",
    "make_boolean": "Stairs.invert",
    "invert": "Stairs.make_boolean",
}

_relationship_map = {
    "lt": "strictly less than",
    "le": "less than, or equal to,",
    "gt": "strictly greater than",
    "ge": "greater than, or equal to,",
    "eq": "equal to",
    "ne": "not equal to",
}

_plots_map = {
    "addition": "s1, s2, s1+s2",
    "subtraction": "s1, s2, s1-s2",
    "multiplication": "s1, s2, s1*s2",
    "division": "s1, s2, s1/(s2+2)",
    "lt": "s1, s2, s1<s2",
    "le": "s1, s2, s1<=s2",
    "gt": "s1, s2, s1>s2",
    "ge": "s1, s2, s1>=s2",
    "eq": "s1, s2, s1==s2",
    "ne": "s1, s2, s1!=s2",
    "and": "s1, s2, s1&s2",
    "or": "s1, s2, s1|s2",
    "invert": "s2, ~s2",
    "make_boolean": "s2, s2.make_boolean()",
    "negate": "s1, -s1",
}

_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> stair_list = [{plots}]
    >>> fig, axes = plt.subplots(nrows=1, ncols={ncols}, figsize=(17,5), sharey=True, sharex=True)
    >>> for ax, title, stair_instance in zip(axes, ({plot_titles}), stair_list):
    ...     stair_instance.plot(ax)
    ...     ax.set_title(title)
"""


def _gen_example(operation):
    plots = _plots_map[operation]
    plot_titles = "[" + ",".join([f'"{x.strip()}"' for x in plots.split(",")]) + "]"
    ncols = len(plots.split(","))

    return _example.format(plots=plots, ncols=ncols, plot_titles=plot_titles)


_arithmetic_binop_docstring_header = """
A binary operator facilitating the {operation} of step functions.
"""

_logical_binop_docstring_header = """
Returns a boolean-valued step function indicating where *self* {operation} *other* are non-zero.
"""

_relational_binop_docstring_header = """
Returns a boolean-valued step function indicating where *self* is {relationship} *other*.
"""

_negate_header = """
An operator which produces a new Stairs instance representing the multiplication of the step function by -1.
"""

_invert_header = """
Returns a boolean-valued step function indicating where *self* is zero-valued.
"""

_common_docstring = """
Equivalent to {op_string}.  See examples below.

Returns
-------
:class:`Stairs`
     new instance representing the result of {op_string}

See Also
--------
{{see_also}}

{{example}}
"""

_common_unop_docstring = _common_docstring.format(op_string=r"{symbol}\ *self*")
_common_binop_docstring = _common_docstring.format(op_string="*self* {symbol} *other*")

_arithmetic_binops = ("addition", "subtraction", "multiplication", "division")
_relational_binops = ("lt", "le", "gt", "ge", "eq", "ne")
_logical_binops = {"and", "or"}


def _get_header(operation):
    if operation in _arithmetic_binops:
        result = _arithmetic_binop_docstring_header.format(operation=operation)
        if operation == "division":
            result = result + "\nThe divisor should cannot be zero-valued anywhere.\n"
        return result
    if operation in _relational_binops:
        relationship = _relationship_map[operation]
        return _relational_binop_docstring_header.format(relationship=relationship)
    if operation in _logical_binops:
        return _logical_binop_docstring_header.format(operation=operation)
    if operation == "negate":
        return _negate_header
    if operation == "invert":
        return _invert_header


def _gen_op_docstring(operation):
    symbol = _symbol_map[operation]
    see_also = _see_also_map[operation]
    example = _gen_example(operation)

    header = _get_header(operation)
    if operation in ("negate", "invert"):
        body = _common_unop_docstring
    else:
        body = _common_binop_docstring
    docstring = header + body
    return docstring.format(symbol=symbol, see_also=see_also, example=example,)


add_docstring = _gen_op_docstring("addition")
subtract_docstring = _gen_op_docstring("subtraction")
multiply_docstring = _gen_op_docstring("multiplication")
divide_docstring = _gen_op_docstring("division")
negate_docstring = _gen_op_docstring("negate")
lt_docstring = _gen_op_docstring("lt")
le_docstring = _gen_op_docstring("le")
gt_docstring = _gen_op_docstring("gt")
ge_docstring = _gen_op_docstring("ge")
eq_docstring = _gen_op_docstring("eq")
ne_docstring = _gen_op_docstring("ne")
logical_and_docstring = _gen_op_docstring("and")
logical_or_docstring = _gen_op_docstring("or")
invert_docstring = _gen_op_docstring("invert")


identical_docstring = """
Returns True if *self* and *other* represent the same step functions.

Returns
-------
boolean

See Also
--------
Stairs.eq, Stairs.ne

Examples
--------

>>> s1.identical(s1)
True
>>> s1.identical(s1.copy())
True
>>> s1.identical(s1.copy().layer(1,2))
False
"""

_make_boolean_docstring_body = """
Returns a boolean-valued step function indicating where *self* is non-zero.

Returns
-------
:class:`Stairs`
    A new instance representing where *self* is non-zero

See Also
--------
Stairs.invert
"""

make_boolean_docstring = _make_boolean_docstring_body + _gen_example("make_boolean")



from staircase.core.stairs import Stairs, sample, resample, aggregate, corr, cov

from staircase.core.stats import hist_from_ecdf

from staircase.core.stairs import _mean as mean
from staircase.core.stairs import _median as median
from staircase.core.stairs import _min as min
from staircase.core.stairs import _max as max

from .test_data import make_test_data


def get_version():
    def get_version_post_py38():
        from importlib.metadata import version

        return version(__name__)

    def get_version_pre_py38():
        from pkg_resources import get_distribution

        return get_distribution(__name__).version

    def default():
        return "unknown"

    for func in (get_version_post_py38, get_version_pre_py38, default):
        try:
            return func()
        except Exception:
            pass


__version__ = get_version()



import numpy as np
from pandas import Timestamp
from staircase.core.stats.distribution import percentile_stairs
from staircase.core.tools.datetimes import (
    _convert_date_to_float,
    _convert_float_to_date,
)
from staircase.util._decorators import Appender
from staircase.core.stats import docstrings
import staircase as sc


@Appender(docstrings.integral_and_mean_example, join="\n", indents=1)
def _get_integral_and_mean(self, lower=float("-inf"), upper=float("inf")):
    """
    Calculates the integral, and the mean of the step function.


    Parameters
    ----------
    lower : int, float or pandas.Timestamp
        lower bound of the interval on which to perform the calculation
    upper : int, float or pandas.Timestamp
        upper bound of the interval on which to perform the calculation

    Returns
    -------
    tuple
        The area and mean are returned as a pair

    See Also
    --------
    Stairs.integrate, Stairs.mean
    """
    new_instance = self.clip(lower, upper)
    if new_instance.number_of_steps() < 2:
        return 0, np.nan
    if lower != float("-inf"):
        new_instance[lower] = new_instance._get(lower, 0)
    if upper != float("inf"):
        new_instance[upper] = new_instance._get(upper, 0)
    cumulative = new_instance._cumulative()
    widths = np.subtract(cumulative.keys()[2:], cumulative.keys()[1:-1])
    heights = cumulative.values()[1:-1]
    area = np.multiply(widths, heights).sum()
    mean = area / (cumulative.keys()[-1] - cumulative.keys()[1])
    return area, mean


@Appender(docstrings.integral_and_mean_example, join="\n", indents=1)
def get_integral_and_mean(self, lower=float("-inf"), upper=float("inf")):
    if isinstance(lower, Timestamp):
        lower = _convert_date_to_float(lower, self.tz)
    if isinstance(upper, Timestamp):
        upper = _convert_date_to_float(upper, self.tz)
    return _get_integral_and_mean(self, lower, upper)


@Appender(docstrings.integrate_example, join="\n", indents=1)
def integrate(self, lower=float("-inf"), upper=float("inf")):
    """
    Calculates the integral of the step function.

    Parameters
    ----------
    lower : int, float or pandas.Timestamp
        lower bound of the interval on which to perform the calculation
    upper : int, float or pandas.Timestamp
        upper bound of the interval on which to perform the calculation

    Returns
    -------
    float
        The area

    See Also
    --------
    Stairs.get_integral_and_mean
    """
    area, _ = get_integral_and_mean(self, lower, upper)
    return area


@Appender(docstrings.mean_example, join="\n", indents=1)
def mean(self, lower=float("-inf"), upper=float("inf")):
    """
    Calculates the mean of the step function.

    Parameters
    ----------
    lower : int, float or pandas.Timestamp
        lower bound of the interval on which to perform the calculation
    upper : int, float or pandas.Timestamp
        upper bound of the interval on which to perform the calculation

    Returns
    -------
    float
        The mean

    See Also
    --------
    Stairs.rolling_mean, Stairs.get_integral_and_mean, Stairs.median, Stairs.mode
    """
    _, mean = get_integral_and_mean(self, lower, upper)
    return mean


@Appender(docstrings.percentile_example, join="\n", indents=1)
def percentile(self, x, lower=float("-inf"), upper=float("inf")):
    """
    Calculates the x-th percentile of the step function's values

    Parameters
    ----------
    lower : int, float or pandas.Timestamp
        lower bound of the interval on which to perform the calculation
    upper : int, float or pandas.Timestamp
        upper bound of the interval on which to perform the calculation

    Returns
    -------
    float
        The x-th percentile

    See Also
    --------
    Stairs.median, Stairs.percentile_stairs
    """
    assert 0 <= x <= 100
    percentiles = percentile_stairs(self, lower, upper)
    return (percentiles(x, how="left") + percentiles(x, how="right")) / 2


@Appender(docstrings.median_example, join="\n", indents=1)
def median(self, lower=float("-inf"), upper=float("inf")):
    """
    Calculates the median of the step function.

    Parameters
    ----------
    lower : int, float or pandas.Timestamp
        lower bound of the interval on which to perform the calculation
    upper : int, float or pandas.Timestamp
        upper bound of the interval on which to perform the calculation

    Returns
    -------
    float
        The median

    See Also
    --------
    Stairs.mean, Stairs.mode, Stairs.percentile, Stairs.percentile_Stairs
    """
    return percentile(self, 50, lower, upper)


@Appender(docstrings.mode_example, join="\n", indents=1)
def mode(self, lower=float("-inf"), upper=float("inf")):
    """
    Calculates the mode of the step function.

    If there is more than one mode only the smallest is returned

    Parameters
    ----------
    lower : int, float or pandas.Timestamp, optional
        lower bound of the interval on which to perform the calculation
    upper : int, float or pandas.Timestamp, optional
        upper bound of the interval on which to perform the calculation

    Returns
    -------
    float
        The mode

    See Also
    --------
    Stairs.mean, Stairs.median
    """
    df = (
        self.clip(lower, upper)
        .to_dataframe()
        .iloc[1:-1]
        .assign(duration=lambda df: df.end - df.start)
    )
    return df.value.loc[df.duration.idxmax()]


@Appender(docstrings.var_example, join="\n", indents=1)
def var(self, lower=float("-inf"), upper=float("inf")):
    """
    Calculates the variance of the step function.

    Parameters
    ----------
    lower : int, float or pandas.Timestamp
        lower bound of the interval on which to perform the calculation
    upper : int, float or pandas.Timestamp
        upper bound of the interval on which to perform the calculation

    Returns
    -------
    float
        The variance of the step function

    See Also
    --------
    Stairs.std
    """
    percentile_minus_mean = (
        percentile_stairs(self, lower, upper) - self.mean(lower, upper)
    )._cumulative()
    return (
        sc.Stairs.from_cumulative(
            dict(
                zip(
                    percentile_minus_mean.keys(),
                    (val * val for val in percentile_minus_mean.values()),
                )
            )
        ).integrate(0, 100)
        / 100
    )


@Appender(docstrings.std_example, join="\n", indents=1)
def std(self, lower=float("-inf"), upper=float("inf")):
    """
    Calculates the standard deviation of the step function.

    Parameters
    ----------
    lower : int, float or pandas.Timestamp
        lower bound of the interval on which to perform the calculation
    upper : int, float or pandas.Timestamp
        upper bound of the interval on which to perform the calculation

    Returns
    -------
    float
        The standard deviation of the step function

    See Also
    --------
    Stairs.var
    """
    return np.sqrt(var(self, lower, upper))



from staircase.core.tools.datetimes import check_binop_timezones
from staircase.core.aggregation import _min_pair, _max_pair
from staircase.util._decorators import Appender
from staircase.core.ops import docstrings
import staircase as sc


@Appender(docstrings.make_boolean_docstring, join="\n", indents=1)
def make_boolean(self):
    new_instance = self != sc.Stairs(0)
    return new_instance


@Appender(docstrings.invert_docstring, join="\n", indents=1)
def invert(self):
    new_instance = make_boolean(self)
    new_instance = sc.Stairs(1, self.use_dates, self.tz) - new_instance
    return new_instance


def _make_logical_func(pair_func, docstring):
    @Appender(docstring, join="\n", indents=1)
    def func(self, other):
        assert isinstance(other, type(self)), "Arguments must be both of type Stairs."
        check_binop_timezones(self, other)
        self_bool = self.make_boolean()
        other_bool = other.make_boolean()
        return pair_func(self_bool, other_bool)

    return func


logical_and = _make_logical_func(_min_pair, docstrings.logical_and_docstring)
logical_or = _make_logical_func(_max_pair, docstrings.logical_or_docstring)



from staircase.core.tools.datetimes import check_binop_timezones
import staircase as sc


def _sanitize_binary_operands(self, other, copy_other=False):
    if not isinstance(other, sc.Stairs):
        other = sc.Stairs(other, self.use_dates, self.tz)
    else:
        check_binop_timezones(self, other)
        if copy_other:
            other = other.copy()
    return self.copy(), other
